/* TestStress.java
 *
 * -----------------------------------------------------------------
 * Licensed Materials - Property of IBM
 * 5724Q99
 * (C) Copyright IBM Corp. 1998, 2006
 * All Rights Reserved
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 * -----------------------------------------------------------------
 * $Id: //depot/rbo/src/packages/ibmee/javaapi/examples/TestStress.java#3 $
 */
package examples;

import java.net.InetAddress;
import java.util.Date;
import java.util.HashMap;

import com.ibm.rbapi.RBActivity;
import com.ibm.rbapi.RBActivityTarget;
import com.ibm.rbapi.RBActivityTargetRecs;
import com.ibm.rbapi.RBActivityTemplate;
import com.ibm.rbapi.RBActivityTemplateRecs;
import com.ibm.rbapi.RBConfiguration;
import com.ibm.rbapi.RBConfigurationRecs;
import com.ibm.rbapi.RBHost;
import com.ibm.rbapi.RBHostRecs;
import com.ibm.rbapi.RBOSDeploymentTemplate;
import com.ibm.rbapi.RBServer;



/*
 * TestStress: put heavy load on one TPMFOSD server
 * Create one thread per client that
 * - repeatedly deploys a configuration
 * - monitors deployment with two RPC calls per second
 * The first exception stops the whole test.
 * 
 * Command line argument: cleanup -> do only cleanup of old templates
 * 
 * Properties:
 * - java.util.logging.config.file
 * - server.ip
 * - server.port
 * - server.secret
 * - client.count
 * - client.maxtime     // max time for a deployment (minutes)
 * - client0.ip
 * - client0.config
 * - client1.ip
 * - client1.config
 * - ...
 */
public class TestStress {

    private RBServer rbServer;
    private boolean finished;

    /*
     * Main program: handle cleanup and exit OR
     * create a TestStress instance that will start multiple threads.
     * Wait for one thread so raise the "finished" flag on the instance.
     * Before leaving, display RPC statistics.
     */
    public static void main(String args[]) {
        try {
            String serverip = System.getProperty("server.ip");
            System.out.println("Will stress server "+serverip);
            InetAddress ip = InetAddress.getByName(serverip);
            int port = Integer.parseInt(System.getProperty("server.port","8080"));
            String secret = System.getProperty("server.secret", "");
            RBServer rbServer = new RBServer(ip, port, secret);
            System.out.println("TestStress: created rbServer\n");
            
            // Cleanup case
            if (args.length>0 && args[0].equals("cleanup")) {
                System.out.println("TestStress: cleanup");
                RBActivityTemplateRecs tr = rbServer.getTemplates("Description like 'Stress%'", "");
                System.out.println("TestStress: found "+tr.size()+" templates to cleanup");
                RBActivityTemplate tpls[] = tr.subList(0, tr.size());
                for (int i=0; i<tpls.length; i++) {
                    rbServer.removeTemplate(tpls[i], true);
                }
                System.out.println("TestStress: finished");
                System.out.println(rbServer.getStatistics());
                return;
            }
            
            // Normal case: create a TestStress instance that will start multiple threads
            // Wait for one thread so raise the "finished" flag on the instance
            TestStress ts = new TestStress(rbServer);
            while (!ts.isFinished()) {
                Thread.sleep(2000);
            }
            
            // Before leaving, display RPC statistics            
            System.out.println(rbServer.getStatistics());
            
            System.out.println("TestStress: finished");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /*
     * Constructor: start one thread (StressClient) per client
     */
    public TestStress(RBServer rbServer)
    {
        this.rbServer = rbServer;
        finished = false;
        int limit = Integer.parseInt(System.getProperty("client.count", "0"));
        long maxtime = Long.parseLong(System.getProperty("client.maxtime", "0")) * 60000;
        System.out.println("TestStress: starting " + limit + " clients...");
        for (int i = 0; i < limit; i++) {
            String cltnum = "client" + i + ".";
            StressClient clt = new StressClient(this,
                    System.getProperty(cltnum + "ip"),
                    System.getProperty(cltnum + "config"),
                    maxtime);
            clt.start();
        }
        System.out.println("TestStress: all clients started.");
    }
    
    public RBServer getRbServer() {
        return rbServer;
    }

    public synchronized boolean isFinished() {
        return finished;
    }

    public synchronized void setFinished(boolean finished) {
        this.finished = finished;
    }

    ////////////////////////////////////////////////////////////////////////////////////////
 
    /*
     * Independent thread that 
     * - repeatedly deploys a configuration
     * - monitors deployment with two RPC calls per second
     * 
     * THIS "POLLING" IS VERY INEFFICIENT AND MUST NOT BE USED WITH MANY CLIENTS
     * A BETTER APPROACH IS ILLUSTRATED IN examples/TestStressSmart.java
     */
    class StressClient extends Thread {

        private TestStress test;
        private String ip;
        private String config;
        private RBServer rbServer;
        private long maxtime;   // millis
        
        public StressClient(TestStress test,String ip,String config,long maxtime) 
        {
            this.test = test;
            this.ip = ip;
            this.config = config;
            this.rbServer = test.getRbServer();
            this.maxtime = maxtime;
        }
        
        public void run() 
        {
            try {
                
                // instantiate "constant" objects (configuration, target, template)
                
                int nbDepl = 0;
                RBConfigurationRecs recs = rbServer.getConfigurations(
                        "[Description]='"+config+"'", "");
                if (recs.size() != 1) {
                    throw new RuntimeException(config + " matches " + recs.size() + " configurations");
                }
                RBConfiguration rbconf = recs.elementAt(0);
                RBHostRecs hosts = rbServer.getHosts("[IP]='"+ip+"'", "");
                RBHost target;
                if (hosts.size()==0) {
                    target = rbServer.registerHost(ip, "", "", "");
                } else {
                    target = hosts.elementAt(0);
                }
                RBOSDeploymentTemplate t = rbServer.registerOSDeploymentTemplate(
                        "Stress "+config+" on "+ip, null);
                
                // repeatedly start and watch one deployment
                
                while (!test.isFinished()) {
                    HashMap paramMap = new HashMap();
                    paramMap.put("Configuration", rbconf);
                    RBActivity a = rbServer.registerActivity("Stress("+nbDepl+")"+ip, 
                            t, false, false, paramMap, target);
                    System.out.println("Client "+ip+": deployment "+nbDepl+" registered");
                    long starttime = new Date().getTime();
                    while (true) {
                        if (test.isFinished()) return;
                        RBActivityTargetRecs tr = a.getActivityTargets();   // 1st RPC call
                        if (test.isFinished()) return;
                        RBActivityTarget at = tr.elementAt(0);              // 2nd RPC call
                        if (test.isFinished()) return;
                        String status = at.getStatus();
                        if (!status.equals("PENDING") && !status.equals("INPROGRESS")) {
                            System.out.println("Client "+ip+": deployment "+nbDepl+" "+status);
                            nbDepl++;
                            break;
                        }
                        Thread.sleep(1000);     // try again after 1 second
                        if (maxtime != 0 && new Date().getTime() - starttime > maxtime) {
                            throw new RuntimeException("Deployment is taking too long!");
                        }
                    }
                }
            } catch (Exception e) {
                System.out.println("Client "+ip+": exception caught");
                e.printStackTrace(System.out);
                System.out.println("Client "+ip+": setting finished flag and exiting");
                test.setFinished(true);
            }
        }
    }
}
