/* TestStressSmart.java
 *
 * -----------------------------------------------------------------
 * Licensed Materials - Property of IBM
 * 5724Q99
 * (C) Copyright IBM Corp. 1998, 2006
 * All Rights Reserved
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 * -----------------------------------------------------------------
 * $Id: //depot/rbo/src/packages/ibmee/javaapi/examples/TestStressSmart.java#2 $
 */
package examples;

import java.net.InetAddress;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

import com.ibm.rbapi.RBActivity;
import com.ibm.rbapi.RBActivityTemplate;
import com.ibm.rbapi.RBActivityTemplateRecs;
import com.ibm.rbapi.RBConfiguration;
import com.ibm.rbapi.RBConfigurationRecs;
import com.ibm.rbapi.RBEvent;
import com.ibm.rbapi.RBHost;
import com.ibm.rbapi.RBHostRecs;
import com.ibm.rbapi.RBOSDeploymentTemplate;
import com.ibm.rbapi.RBServer;

/*
 * TestStressSmart: put heavy load on one TPMFOSD server
 * but with a smarter monitoring of events than TestStress.
 * Create one thread per client that
 * - repeatedly deploys a configuration
 * - monitors deployment with an efficient "delegated" approach
 * The first exception stops the whole test.
 * 
 * Command line argument: cleanup -> do only cleanup of old templates
 * 
 * Properties:
 * - java.util.logging.config.file
 * - server.ip
 * - server.port
 * - server.secret
 * - client.count
 * - client0.ip
 * - client0.config
 * - client1.ip
 * - client1.config
 * - ...
 */
public class TestStressSmart {

    private RBServer rbServer;
    private boolean finished;
    private SimpleEventMonitor evtMon;

    /*
     * Main program: handle cleanup and exit OR
     * create a TestStressSmart instance that will start multiple threads.
     * Wait for one thread so raise the "finished" flag on the instance.
     * Before leaving, display RPC statistics.
     */
    public static void main(String args[]) {
        try {
            String serverip = System.getProperty("server.ip");
            System.out.println("Will stress server "+serverip);
            InetAddress ip = InetAddress.getByName(serverip);
            int port = Integer.parseInt(System.getProperty("server.port","8080"));
            String secret = System.getProperty("server.secret", "");
            RBServer rbServer = new RBServer(ip, port, secret);
            System.out.println("TestStressSmart: created rbServer\n");
            
            // Cleanup case
            if (args.length>0 && args[0].equals("cleanup")) {
                System.out.println("TestStressSmart: cleanup");
                RBActivityTemplateRecs tr = rbServer.getTemplates("Description like 'Smart%'", "");
                System.out.println("TestStressSmart: found "+tr.size()+" templates to cleanup");
                RBActivityTemplate tpls[] = tr.subList(0, tr.size());
                for (int i=0; i<tpls.length; i++) {
                    rbServer.removeTemplate(tpls[i], true);
                }
                System.out.println("TestStressSmart: finished");
                System.out.println(rbServer.getStatistics());
                return;
            }
            
            // Normal case: create a TestStressSmart instance that will start multiple threads
            // Wait for one thread so raise the "finished" flag on the instance
            TestStressSmart ts = new TestStressSmart(rbServer);
            while (!ts.isFinished()) {
                Thread.sleep(2000);
            }
            
            // Before leaving, display RPC statistics            
            System.out.println(rbServer.getStatistics());
            
            System.out.println("TestStressSmart: finished");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /*
     * Constructor: start event monitoring thread plus one thread (StressClient) per client
     */
    public TestStressSmart(RBServer rbServer)
    {
        this.rbServer = rbServer;
        finished = false;
        evtMon = new SimpleEventMonitor(this);
        evtMon.start();
        int limit = Integer.parseInt(System.getProperty("client.count", "0"));
        System.out.println("TestStressSmart: starting " + limit + " clients...");
        for (int i = 0; i < limit; i++) {
            String cltnum = "client" + i + ".";
            StressClient clt = new StressClient(this,
                    System.getProperty(cltnum + "ip"),
                    System.getProperty(cltnum + "config"));
            clt.start();
        }
        System.out.println("TestStressSmart: all clients started.");
    }
    
    public RBServer getRbServer() {
        return rbServer;
    }

    public synchronized boolean isFinished() {
        return finished;
    }

    public synchronized void setFinished(boolean finished) {
        this.finished = finished;
    }
    
    public SimpleEventMonitor getEvtMon() {
        return evtMon;
    }

    ////////////////////////////////////////////////////////////////////////////////////////
 
    /*
     * Independent thread that 
     * - repeatedly deploys a configuration
     * - delegates event monitoring to an external thread
     * 
     * THIS IS A MUCH BETTER APPROACH FOR MANY CLIENTS THAN THE ONE
     * USED IN examples/TestStress.java
     */
    class StressClient extends Thread {

        private TestStressSmart test;
        private String ip;
        private String config;
        private RBServer rbServer;
        
        public StressClient(TestStressSmart test,String ip,String config) 
        {
            this.test = test;
            this.ip = ip;
            this.config = config;
            this.rbServer = test.getRbServer();
        }
        
        public void run() 
        {
            try {
                
                // instantiate "constant" objects (configuration, target, template)
                
                int nbDepl = 0;
                RBConfigurationRecs recs = rbServer.getConfigurations(
                        "[Description]='"+config+"'", "");
                if (recs.size() != 1) {
                    throw new RuntimeException(config + " matches " + recs.size() 
                            + " configurations");
                }
                RBConfiguration rbconf = recs.elementAt(0);
                RBHostRecs hosts = rbServer.getHosts("[IP]='"+ip+"'", "");
                RBHost target;
                if (hosts.size()==0) {
                    target = rbServer.registerHost(ip, "", "", "");
                } else {
                    target = hosts.elementAt(0);
                }
                RBOSDeploymentTemplate t = rbServer.registerOSDeploymentTemplate(
                        "Smart "+config+" on "+ip, null);
                
                // repeatedly start and watch one deployment
                
                while (!test.isFinished()) {
                    HashMap paramMap = new HashMap();
                    paramMap.put("Configuration", rbconf);
                    RBActivity a = rbServer.registerActivity("Stress("+nbDepl+")"+ip, 
                            t, false, false, paramMap, target);
                    System.out.println("Client "+ip+": deployment "+nbDepl+" registered");
                    
                    // Delegate progress tracking for the new activity.
                    // Note that there is a little synchronization risk here: 
                    // If the activity starts very quickly, evtMon may see event 
                    // and ignore it because we haven't registered yet.
                    // Tracking by RBHost may solve this problem.
                    
                    Long token = test.getEvtMon().delegate(this, a);
                    String status;
                    do {
                        // go to sleep, a signal will be received for events
                        try { sleep(2000); } catch (InterruptedException e) { }
                        if (test.isFinished()) return;
                        // use the token to check type of event
                        RBEvent evt = test.getEvtMon().getEvent(token);
                        if (evt != null) {
                            status = evt.getStatus();
                            if (!status.equals("PENDING")&&!status.equals("INPROGRESS")) {
                                break;
                            } 
                        }
                    } while (true) ;
                    System.out.println("Client "+ip+": deployment "+nbDepl+" "+status);
                    nbDepl++;
                }
            } catch (Exception e) {
                System.out.println("Client "+ip+": exception caught");
                e.printStackTrace(System.out);
                System.out.println("Client "+ip+": setting finished flag and exiting");
                test.setFinished(true);
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////////////////
    
    /*
     * Independent thread that waits for events and notifies waiting theads 
     */
    class SimpleEventMonitor extends Thread {
        
        private TestStressSmart test;
        private HashMap activities;     // maps activities to tokens
        private HashMap callers;        // maps tokens to callers
        private HashMap events;         // maps tokens to events;
        private long lasttoken;
        
        public SimpleEventMonitor(TestStressSmart test) {
            this.test = test;
            activities = new HashMap();
            callers = new HashMap();
            events = new HashMap();
            lasttoken = 0;
        }
        
        /*
         * Get events from server every 5 sec and notify waiting threads
         * Notify all threads if test is finished
         */
        public void run() {
            try {
                Date lastTime = new Date();
                while (true) {
                    RBEvent newevents[] = test.getRbServer().getEvents(lastTime);
                    if (test.isFinished()) return;
                    for (int i=0; i<newevents.length; i++) {
                        RBEvent cur = newevents[i];
                        lastTime = cur.getUpdateTime(); // Update last events found time.
                        synchronized (this) { // prevent changes in activities and callers
                            // find the activity without asking the server
                            RBActivity curact = null;
                            Iterator act = activities.keySet().iterator();
                            while (act.hasNext()) {
                                RBActivity candidate = (RBActivity)act.next();
                                if (cur.belongsTo(candidate)) {
                                    curact = candidate;
                                    break;
                                }
                            }
                            if (curact == null) {
                                continue; // event is not registered, skip it
                            }
                            Long token = (Long)activities.get(curact);
                            // save the event, this must be done before interrupting the caller
                            events.put(token,cur);
                            Thread t = (Thread)callers.get(token);
                            t.interrupt();
                        }
                        if (test.isFinished()) return;
                    }
                    try { Thread.sleep(5000); } catch (InterruptedException e) {} // try again after 5 seconds
                    if (test.isFinished()) return;
                }
            } catch (Exception e) {
                System.out.println("SimpleEventMonitor: exception caught");
                e.printStackTrace(System.out);
                System.out.println("SimpleEventMonitor: setting finished flag and exiting");
                test.setFinished(true);
            }
        }
              
        /*
         * Notify thread t when an event occurs for activity a
         * Save the event with the returned "token"
         * Accept only one caller at a time to preserve integrity 
         */
        public synchronized Long delegate(Thread t, RBActivity a) {
            Long token = new Long(lasttoken++);
            activities.put(a, token);
            callers.put(token, t);
            return token;
        }
        
        /*
         * Return saved event
         */
        public RBEvent getEvent(Long token) {
            return (RBEvent)events.get(token);
        }
    }
    
}
